#ifndef AWS_AUTH_SIGV4_HTTP_REQUEST_H
#define AWS_AUTH_SIGV4_HTTP_REQUEST_H

/**
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0.
 */

#include <aws/auth/auth.h>

AWS_EXTERN_C_BEGIN
AWS_EXTERN_C_END

#endif /* AWS_AUTH_SIGV4_HTTP_REQUEST_H */
