/**
 * Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
 * SPDX-License-Identifier: Apache-2.0.
 */

#ifndef HEADER
#error "Macro HEADER(index, name) must be defined before including this header file!"
#endif

#ifndef HEADER_WITH_VALUE
#error "Macro HEADER_WITH_VALUE(index, name, value) must be defined before including this header file!"
#endif

HEADER(1, ":authority")
HEADER_WITH_VALUE(2, ":method", "GET")
HEADER_WITH_VALUE(3, ":method", "POST")
HEADER_WITH_VALUE(4, ":path", "/")
HEADER_WITH_VALUE(5, ":path", "/index.html")
HEADER_WITH_VALUE(6, ":scheme", "http")
HEADER_WITH_VALUE(7, ":scheme", "https")
HEADER_WITH_VALUE(8, ":status", "200")
HEADER_WITH_VALUE(9, ":status", "204")
HEADER_WITH_VALUE(10, ":status", "206")
HEADER_WITH_VALUE(11, ":status", "304")
HEADER_WITH_VALUE(12, ":status", "400")
HEADER_WITH_VALUE(13, ":status", "404")
HEADER_WITH_VALUE(14, ":status", "500")
HEADER(15, "accept-charset")
HEADER_WITH_VALUE(16, "accept-encoding", "gzip,deflate")
HEADER(17, "accept-language")
HEADER(18, "accept-ranges")
HEADER(19, "accept")
HEADER(20, "access-control-allow-origin")
HEADER(21, "age")
HEADER(22, "allow")
HEADER(23, "authorization")
HEADER(24, "cache-control")
HEADER(25, "content-disposition")
HEADER(26, "content-encoding")
HEADER(27, "content-language")
HEADER(28, "content-length")
HEADER(29, "content-location")
HEADER(30, "content-range")
HEADER(31, "content-type")
HEADER(32, "cookie")
HEADER(33, "date")
HEADER(34, "etag")
HEADER(35, "expect")
HEADER(36, "expires")
HEADER(37, "from")
HEADER(38, "host")
HEADER(39, "if-match")
HEADER(40, "if-modified-since")
HEADER(41, "if-none-match")
HEADER(42, "if-range")
HEADER(43, "if-unmodified-since")
HEADER(44, "last-modified")
HEADER(45, "link")
HEADER(46, "location")
HEADER(47, "max-forwards")
HEADER(48, "proxy-authenticate")
HEADER(49, "proxy-authorization")
HEADER(50, "range")
HEADER(51, "referer")
HEADER(52, "refresh")
HEADER(53, "retry-after")
HEADER(54, "server")
HEADER(55, "set-cookie")
HEADER(56, "strict-transport-security")
HEADER(57, "transfer-encoding")
HEADER(58, "user-agent")
HEADER(59, "vary")
HEADER(60, "via")
HEADER(61, "www-authenticate")
